/*
 * Copyright (C) 2017-2023 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef LOGGER_HPP_
#define LOGGER_HPP_

#include "compiler.hpp"

#include <print.hpp>

#include <mutex>
#include <sstream>
#include <syslog.h>

namespace server {
extern unsigned logging;
extern std::mutex loglock;
extern void logger(const std::string& msg);

template<class... Args>
void debug(unsigned level, std::string_view fmt, const Args&... args) {
    if(level <= logging) {
        try {
            if(fmt.back() == '\n')
                fmt.remove_suffix(1);
            auto msg = print_format(fmt, args...);
            const std::lock_guard lock(loglock);
            print(std::cerr, "debug: {}\n", msg);
            logger(msg);
        } catch(...) {}
    }
}

template<class... Args>
void log_info(std::string_view fmt, const Args&... args) {
    if(fmt.back() == '\n')
        fmt.remove_suffix(1);
    auto msg = print_format(fmt, args...);
    const std::lock_guard lock(loglock);
    ::syslog(LOG_INFO, "%s", msg.c_str());
    logger(msg);
    if(logging > 1)
        print(std::cerr, "info: {}\n", msg);
}

template<class... Args>
void log_notice(std::string_view fmt, const Args&... args) {
    if(fmt.back() == '\n')
        fmt.remove_suffix(1);
    auto msg = print_format(fmt, args...);
    const std::lock_guard lock(loglock);
    ::syslog(LOG_NOTICE, "%s", msg.c_str());
    logger(msg);
    if(logging > 0)
        print(std::cerr, "notice: {}\n", msg);
}

template<class... Args>
void log_warn(std::string_view fmt, const Args&... args) {
    if(fmt.back() == '\n')
        fmt.remove_suffix(1);
    auto msg = print_format(fmt, args...);
    const std::lock_guard lock(loglock);
    ::syslog(LOG_WARNING, "%s", msg.c_str());
    logger(msg);
    if(logging)
        print(std::cerr, "warn: {}\n", msg);
}

template<class... Args>
void log_error(std::string_view fmt, const Args&... args) {
    if(fmt.back() == '\n')
        fmt.remove_suffix(1);
    auto msg = print_format(fmt, args...);
    const std::lock_guard lock(loglock);
    ::syslog(LOG_ERR, "%s", msg.c_str());
    logger(msg);
    if(logging)
        print(std::cerr, "error: {}\n", msg);
}

template<class... Args>
void log_fail(int excode, std::string_view fmt, const Args&... args) {
    if(fmt.back() == '\n')
        fmt.remove_suffix(1);
    auto msg = print_format(fmt, args...);
    const std::lock_guard lock(loglock);
    ::syslog(LOG_CRIT, "%s", msg.c_str());
    logger(msg);
    if(logging)
        print(std::cerr, "fail: {}\n", msg);
    ::exit(excode);
}

inline void log_open(const char *id, int level = LOG_INFO, int facility = LOG_DAEMON, int flags = LOG_CONS | LOG_NDELAY) {
    ::openlog(id, flags, facility);
    ::setlogmask(LOG_UPTO(level));
}

inline void log_close() {
    ::closelog();
}
} // end namespace server

/*!
 * Server logging and output.  This makes it easy to write
 * output and fifo control messages using std::format strings.
 * \file logger.hpp
 */
#endif
