/*
 * Copyright (C) 2020-2023 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "main.hpp"
#include "events.hpp"
#include "logger.hpp"
#include "network.hpp"
#include "features.hpp"
#include "messages.hpp"
#include "registry.hpp"
#include "calls.hpp"
#include "ipc.hpp"

#include <args.hpp>
#include <strings.hpp>
#include <keyfile.hpp>
#include <filesystem.hpp>

#include <thread>
#include <csignal>
#include <climits>
#include <cstdio>

#include <unistd.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <pwd.h>
#include <mqueue.h>

#ifdef  SYSTEMD_SERVICE
#include <systemd/sd-daemon.h>
#else
inline auto sd_notify(int, const char *) {  // NOLINT
    return 0;
}
#endif

using namespace server;

// some global flags, easier to just do as external
std::atomic<bool> server::running = false;      // are we running?
std::atomic<bool> server::exiting = false;      // are we exiting?
std::mutex server::loglock;                     // lock logging
unsigned server::logging = 0;                   // verbose logging level?
bool server::ipv6 = false;                      // are we ipv6?
std::string status = "READY=1\nSTATUS=unknown"; // set on reload

namespace { // private data and functions
const auto AGENT_ID = SERVER_ID "/" PROJECT_VERSION;
const char *config_file = SERVER_CONFIG_FILE;
const char *prefix_path = SERVER_PREFIX_PATH;

args::option opt_ipv6('6',  "--ipv6", "enable ipv6");
args::option opt_config('c',"--config", "configuration file", "file");
args::option helpflag('h',  "--help", "display this list");
args::option opt_prefix(0,  "--prefix", "prefix directory", "path");
args::option verbose('v',   "--verbose", "verbosity, repeat for debug", true);
args::option version(0,     "--version", "display version information");
args::option althelp('?',   nullptr, nullptr);

std::string shutdown_message = "Shutting down...";
exosip::context_t udp(AGENT_ID);
exosip::context_t tcp(AGENT_ID);

auto reload() {
    keyfile config;

    // Default server info
    config["server"];
    config["server"]["sitename"] = "Coventry Server";
    config["server"]["location"] = "unspecified";
    config["server"]["webserver"] = "80";
    config["server"]["urlschema"] = "http";

    // Default messages
    config["messages"];
    config["messages"]["welcome"] = "Welcome to Coventry";
    config["messages"]["shutdown"] = "Shutting down...";

    // initialize [calls] with defaults if not set
    config["calls"];
    config["calls"]["mode"] = "proxy";
    config["calls"]["ring"] = "4";
    config["calls"]["delayed"] = "12";

    // initialize [common] with defaults if not set
    config["common"];
    config["common"]["lines"] = "1";
    config["common"]["presence"] = "here";
    config["common"]["type"] = "generic";
    config["common"]["location"] = "unspecified";
    config["common"]["cabling"] = "unspecified";

    // initialize [groups] and [access] so no error occurs
    config["display"];
    config["groups"];
    config["access"];
    config["rooms"];
    config["zones"];

    // load composite config
    config.load(config_file);
    config.load("dynamic.conf");
    config.load("custom.conf");
    config.load("state.conf");

    // process state symlink
    std::string state = "default";
    if(fsys::exists("state.conf") && fsys::is_symlink("state.conf")) {
        auto symlink = fsys::read_symlink("state.conf");
        if(symlink.filename() == symlink)
            state = symlink.stem().string();
    }

    // set run state
    config["server"]["state"] = state;
    debug(3, "current state {}", state);
    status = "READY=1\nSTATUS=" + state;

    // validate config with env overrides
    auto& hostname = config["server"]["hostname"];
    auto& realm = config["server"]["realm"];
    auto& algorithm = config["server"]["algorithm"];
    auto& expires = config["server"]["expires"];
    auto& msg_expires = config["messages"]["expires"];
    auto& names = config["server"]["names"];
    auto& adapter = config["remote"]["address"];
    auto& address = config["server"]["address"];
    auto& port = config["server"]["port"];
    auto& adport = config["remote"]["port"];
    auto& upstream = config["remote"]["upstream"];
    auto& priority = config["server"]["priority"];

    auto family = lower_case(config["server"]["family"]);
    if(family == "af_inet6" || family == "inet6" || family == "ipv6" || family == "6")
        ipv6 = true;

    std::string uri;
    auto env = util::get_env("COVENTRY_SERVER", 255);
    if(env)
        uri = env;

    if(!uri.empty() && uri.substr(0, 4) == "sip:") {
        uri = uri.substr(4);
        auto pos = uri.find_last_of(':');
        if(pos != std::string::npos) {
            port = uri.substr(pos + 1);
            address = uri.substr(0, pos);
        }
        else {
            port = "5060";
            address = uri;
        }
    }

    env = util::get_env("COVENTRY_UPSTREAM", 255);
    if(env)
        upstream = env;

    env = util::get_env("COVENTRY_REMOTE", 255);
    if(env)
        uri = env;
    else
        uri = "";

    if(!uri.empty() && uri.substr(0, 4) == "sip:") {
        uri = uri.substr(4);
        auto pos = uri.find_last_of(':');
        if(pos != std::string::npos) {
            adport = uri.substr(pos + 1);
            adapter = uri.substr(0, pos);
        }
        else {
            adport = "5062";
            adapter = uri;
        }
    }

    if(!upstream.empty()) {
        if(upstream.substr(0, 4) != "sip:")
            log_fail(98, "{}: invalid upstream uri", upstream);
    }

    if(address.empty())
        address = "*";

    if(adapter.empty())
        adapter = "*";

    try {
        address = net::if_address(address);
    }
    catch(const std::exception&) {
        log_fail(98, "{}: binding address invalid", address);
    }

    try {
        adapter = net::if_address(adapter);
    }
    catch(const std::exception&) {
        log_fail(98, "{}: binding address invalid", adapter);
    }

    try {
        (void)stoi(port);
    }
    catch(...) {
        port = "5060";
    }

    try {
        (void)stoi(adport);
    }
    catch(...) {
        adport = "5062";
    }

    try {
        (void)stoi(priority);
    }
    catch(...) {
        priority = "0";
    }

    try {
        (void)stoi(expires);
    }
    catch(...) {
        expires = "300";
    }

    try {
        (void)stoi(msg_expires);
    }
    catch(...) {
        msg_expires = "300";
    }

    if(algorithm.empty())
        algorithm = "SHA-256";

    algorithm = upper_case(algorithm);
    if(algorithm != "MD5" && algorithm != "SHA-256") {
        print(std::cerr, "*** warning: {}: invalid\n", algorithm);
        algorithm = "SHA-256";
    }

    if(hostname.empty()) {
        env = util::get_env("COVENTRY_HOSTNAME", 255);
        if(!env || !*env)
            env = util::get_env("PUBLIC_HOSTNAME", 255);

        if(env && *env)
            hostname = env;
        else
            hostname = net::hostname();
    }
    if(realm.empty())
        realm = hostname;
    net::fqdn(hostname);

    // alternate names we might answer for...
    if(names.empty()) {
        if(hostname != "localhost" && hostname != "localhost.localdomain") {
            auto pos = hostname.find_first_of('.');
            names = hostname.substr(0, pos) + ".local";
            if(pos != std::string::npos)
                names += "," + hostname.substr(0, pos);
            names += ",localhost,localhost.localdomain";
        }
        else if(hostname == "localhost.localdomain")
            names = "localhost";
    }
    names = hostname + "," + names;

    if(!config.exists("features")) {
        // Default features if no feature plan...
        config["features"];
        config["features"]["*99"] = "echo";
        config["features"]["*98"] = "reload";
        config["features"]["*97"] = "@weather";
    }

    running = true;
    return config;
}

void msgs(const pbx_msg& msg) {
    if(msg.ver != PBX_VERSION) {
        log_error("ipc version invalid");
        return;
    }
    switch(msg.type) {
    case PBX_LEVEL:
        change_level(msg.body.level);
        break;
    case PBX_MESSAGE:
        exec::message_system(msg);
        break;
    case PBX_RELOAD:
        exec::message_system(msg);
        raise(SIGHUP);
        break;
    case PBX_ACTIVE:
    case PBX_ASSIGN:
    case PBX_CLEARS:
    case PBX_PRESENCE:
        exec::pbx_action(msg);
        break;
    default:
        debug(4, "received unknown message {}", static_cast<int>(msg.type));
    }
}

void stop() {
    sd_notify(1, "STOPPING=1");
    running = false;
    system_message(shutdown_message);
    exec::stop_registry();
}
} // anon namespace

auto main(int argc, const char **argv) -> int { // NOLINT
    auto exit_code = 0;
    atexit([]{
        if(running) {
            stop();
            log_warn("exiting server");
        }
        mq_unlink(SERVER_IPCPATH);
        shm_unlink(SERVER_IPCPATH ".registry");
        shm_unlink(SERVER_IPCPATH ".calls");
    });

    try {
        auto count = args::parse(argc, argv);
        if(is(helpflag) || is(althelp) || count) {
            args::help({SERVER_ID " [options]"}, {
                "coventry " PROJECT_VERSION,
                "David Sugar <tychosoft@gmail.com>",
                "Start Coventry Service"
            });
            return 0;
        }

        if(is(opt_config))
            config_file = *opt_config;

        if(is(opt_prefix))
            prefix_path = *opt_prefix;

        if(is(version)) {
            print("Version: {}\nConfig: {}\nPrefix: {}\n", PROJECT_VERSION, config_file, prefix_path);
            return 0;
        }

        if(is(opt_ipv6))
            ipv6 = true;

        umask(007);     // FlawFinder: group access even for root
        fsys::current_path(prefix_path);

        // create ipc info
        fsys::remove("ipc.json");
        std::ofstream ipc("ipc.json");
        ipc << R"({"ipc_path":")" << SERVER_IPCPATH << R"(",)" << std::endl
            << R"( "msg_size":)" << sizeof(pbx_msg) << "," << std::endl
            << R"( "sys_size":)" << sizeof(pbx_sys) << "," << std::endl
            << R"( "reg_size":)" << sizeof(pbx_reg) << "," << std::endl
            << R"( "call_size":)" << sizeof(pbx_call) << "," << std::endl
            << R"( "reg_count":)" << registry_entries << "," << std::endl
            << R"( "call_count":)" << SERVER_CALL_LIMIT << std::endl
            << "}" << std::endl;
        ipc.close();

        // start mqueue
        struct mq_attr attr{}, old_attr{};
        memset(&attr, 0, sizeof(attr));
        attr.mq_maxmsg = 10;
        attr.mq_msgsize = sizeof(pbx_msg);
        auto mq = mq_open (SERVER_IPCPATH, O_RDWR | O_CREAT, 0660, &attr);
        if(mq == MQD_INVALID)
            throw std::runtime_error(print_format("message queue failure {}", errno));

        // setup logging
        logging = verbose.count();
        log_open(SERVER_ID, LOG_INFO);

        std::set_new_handler([]() {
            die(9, "*** coventry: out of memory\n");
        });

        // protect signals from other threads
        sigset_t sigs;
        sigemptyset(&sigs);
        sigaddset(&sigs, SIGALRM);
        sigaddset(&sigs, SIGINT);
        sigaddset(&sigs, SIGHUP);
        sigaddset(&sigs, SIGKILL);
        sigaddset(&sigs, SIGTERM);
        sigaddset(&sigs, SIGUSR1);      // mq events
        pthread_sigmask(SIG_BLOCK, &sigs, nullptr);

        // start services
        auto config = reload();
        shutdown_message = config.at("messages").at("shutdown");

        // start threads
        std::thread calls(exec_calls, config);
        std::thread features(exec_features, config);
        std::thread messages(exec_messages, config);
        std::thread registry(exec_registry, config);
        std::thread udp_thread(exec_local, std::ref(udp), config);
        std::thread tcp_thread(exec_remote, std::ref(tcp), config);
        std::this_thread::sleep_for(std::chrono::milliseconds(20));

        // system notifiers
        struct sigevent mq_event{};
        mq_event.sigev_notify = SIGEV_SIGNAL;
        mq_event.sigev_signo = SIGUSR1;
        mq_notify(mq, &mq_event);
        attr.mq_flags = O_NONBLOCK;
        mq_setattr(mq, &attr, &old_attr);

        // server main loop and signal handling, call stop() to exit
        auto pri = 0U;
        pbx_msg msg{};

        auto timer = sync_clock() + std::chrono::seconds(60);
        alarm(60);
        log_info("started server");
        sd_notify(0, status.c_str());
        while(running) {
            int sig{};
            if(sigwait(&sigs, &sig))
                break;
            switch(sig) {
            case SIGINT:
                std::cout << std::endl;  // for c/r after ^C
                stop();
                break;
            case SIGALRM:
                // sync alarm every 60 seconds, used for stats...
                sd_notify(0, "WATCHDOG=1");
                debug(3, "minute timer wakeup");
                timer += std::chrono::seconds(60);
                alarm(std::chrono::duration_cast<std::chrono::seconds>(timer - sync_clock()).count());
                break;
            case SIGTERM:
            case SIGQUIT:
                stop();
                break;
            case SIGHUP:
                sd_notify(0, "RELOADING=1");
                log_info("reloading server");
                config = reload();
                exec::reload_registry(config);
                exec::reload_messages(config);
                exec::reload_features(config);
                exec::reload_calls(config);
                sd_notify(0, status.c_str());
                break;
            case SIGUSR1:
                while(mq_receive(mq, reinterpret_cast<char *>(&msg), sizeof(msg), &pri) != -1)
                    msgs(msg);
                mq_notify(mq, &mq_event);
                break;
            default:
                break;
            }
        }

        // assuming we hit stop(), bring system down cleanly, join our threads
        registry.join();
        messages.join();
        features.join();
        calls.join();

        exiting = true;
        udp.close();
        if(tcp.active())
            tcp.close();

        udp_thread.join();
        tcp_thread.join();
    }
    catch(const bad_arg& e) {
        die(-1, "Usage: {} [options]\nerror: {}\n", SERVER_ID, e.what());
    }
    catch(const std::exception& e) {
        if(running)
            log_fail(-1, "{}", e.what());
        else
            die(-1, "error: {}\n", e.what());
    }
    catch(...) {
        if(running)
            log_fail(-1, "unknown failure");
        else
            die(-1, "error: unknown failure\n");
    }

    log_info("stopped server; reason={}", exit_code);
    return exit_code;
}
