/*
 * Copyright (C) 2020-2023 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef	REGISTRY_HPP_
#define REGISTRY_HPP_

#include "compiler.hpp"
#include "nonce.hpp"
#include "ipc.hpp"
#include "exosip.hpp"

#include <stdexcept>
#include <array>
#include <functional>
#include <vector>
#include <unordered_map>
#include <unordered_set>
#include <string>
#include <cstdint>
#include <atomic>
#include <mutex>
#include <ctime>

inline constexpr auto registry_entries = 80U;
inline constexpr auto dialplan_min = 10U;
inline constexpr auto dialplan_max = dialplan_min + registry_entries - 1U;
inline constexpr auto expires_year = 60L * 60L * 24L * 365L;

using pbx_var = const std::shared_ptr<pbx_msg>;

namespace server {
class event_t;
class registry;
using keys = std::unordered_map<std::string, std::string>;

class contact_t final {
public:
    inline contact_t() = default;
    explicit contact_t(const osip_uri_t *uri);

    inline explicit contact_t(const event_t& event) {
        from(event);
    }

    inline explicit operator bool() const {
        return !host_.empty();
    }

    inline auto operator!() const {
        return host_.empty();
    }

    inline auto operator==(const contact_t& other) const {
        return host_ == other.host_ && port_ == other.port_ && user_ == other.user_;
    }

    inline auto operator!=(const contact_t& other) const {
        return host_ != other.host_ || port_ != other.port_ || user_ != other.user_;
    }

    inline auto port() const {return port_;}
    inline auto user() const {return user_;}
    inline auto host() const {return host_;}
    inline auto display() const {return display_;}

    inline auto empty() const {
        return host_.empty();
    }

    inline auto id() const {
        auto result = host_;
        if(host_.empty())
            return result;
        if(port_)
            result += ":" + std::to_string(port_);
        if(user_.empty())
            return result;
        return user_ + "@" + result;
    }

    inline auto addr() const {
        auto result = host_;
        if(host_.empty())
            return result;
        if(port_)
            result += ":" + std::to_string(port_);
        return result;
    }

    void from(const event_t& event);

protected:
    friend class registry;

    std::string host_;
    uint16_t port_{};
    std::string user_;
    std::string display_;
};

class registry final {
public:
    inline registry() : ptr_(nullptr) {}
    explicit registry(unsigned key);
    registry(const registry&) = default;

    inline explicit operator eXosip_t *() const {
        return context_;
    }

    explicit operator bool() const;
    auto operator!() const -> bool;

    inline auto operator==(const registry& from) {
        return get() == from.get();
    }

    inline auto presence() const {
        if(!shm_ && operator bool())
            return pbx_reg::HERE;
        if(!shm_)
            return pbx_reg::GONE;
        return shm_->presence;
    }

    inline auto open() const {  // FlawFinder: useless
        return operator bool() && (presence() == pbx_reg::HERE || presence() == pbx_reg::AWAY);
    }

    inline auto extension() const {
        return ptr_ ? ptr_->extension : 0;
    }

    inline auto value() const {
        return ptr_ ? ptr_->extension : 0;
    }

    inline auto expires() const {
        return ptr_ ? ptr_->expires.load() : 0;
    }

    inline auto schema() const {
        return ptr_ ? ptr_->schema.c_str() : "";
    }

    inline auto context() const {
        return context_;
    }

    inline auto hash() const {
        return static_cast<size_t>(value());
    }

    inline auto use_count() const {
        return ptr_.use_count();
    }

    inline auto empty() const {
        return !ptr_;
    }

    inline auto nonce() const {
        return ptr_ ? ptr_->nonce : nonce_t();
    }

    inline auto allow_invite() const {
        return shm_ ? shm_->flags.invite && open() : false;   // FlawFinder: no
    }

    inline auto allow_coverage() const {
        return shm_ ? shm_->flags.invite && operator bool() && presence() == pbx_reg::HERE : false;
    }

    inline auto allow_message() const {
        return shm_ ? shm_->flags.message && open() : false;  // FlawFinder: no
    }

    inline auto welcome() const {
        auto result = false;
        if(ptr_ && shm_) {
            result = shm_->flags.message && ptr_->welcome;
            ptr_->welcome = false;
        }
        return result;
    }

    inline auto user_id() const {
        return ptr_ ? ptr_->contact.user() : "";
    }

    inline auto display() const {
        return ptr_ ? ptr_->contact.display() : "";
    }

    inline auto remote() const {
        return ptr_ ? origin().addr() : "";
    }

    inline auto agent() const {
        return ptr_ && context_ ? ptr_-> agent : "unknown";
    }

    inline auto type() const {
        return ptr_ && context_ ? ptr_-> type : "unknown";
    }

    inline auto acl() const {
        return ptr_ && context_ ? ptr_-> acl : std::vector<std::string>();
    }

    inline auto location() const {
        return ptr_ && context_ ? ptr_-> location : "unknown";
    }

    inline auto shm() const {
        return shm_;
    }

    auto local_uri() const -> std::string;
    auto local_alias(std::string_view to) const -> std::string;
    auto local_to() const -> std::string;
    auto origin_uri() const -> std::string;
    auto route_to() const -> std::string;
    auto from(const event_t& event) const -> std::string;
    auto to(const event_t& event) const -> std::string;
    auto origin() const -> contact_t;
    auto contact() const -> contact_t;
    auto keys() const -> server::keys;
    void keys(const server::keys& pairs);
    auto refresh(const event_t& event, time_t period) -> bool;
    void receipt(const event_t& event, time_t period);
    auto verify(const event_t& event) -> bool;
    void drop();
    void allow(exosip::msg_t msg);

    static auto is_extension(std::string_view text) -> bool;

private:
    struct data_t final {
        std::mutex mutex;
        server::keys keys;

        std::atomic<time_t> expires;
        unsigned extension;
        contact_t contact, origin;
        std::string schema;
        nonce_t nonce;
        std::string client;
        std::string agent;
        std::string type;
        std::string location;
        std::vector<std::string> acl;
        uint32_t count;
        bool welcome;
    };

    std::shared_ptr<data_t> ptr_;
    eXosip_t *context_{nullptr};
    pbx_reg *shm_{nullptr};

    inline auto get() const -> data_t * {
        return ptr_.get();
    }
};

void change_level(unsigned level);
void system_message(std::string_view text, std::string_view to = "all", std::string_view subject = "");
void exec_registry(const keyfile& config);

// registry component dispatched actions
namespace exec {
void stop_registry();
void reload_registry(const keyfile& config);
void message_system(const pbx_msg& msg);
void pbx_action(const pbx_msg& ipc);
void feature_code(const event_t& event);
void release_user(const event_t& event);
void refresh_user(const event_t& event);
void option_user(const event_t& event);
void invite_user(const event_t& event);
void message_user(const event_t& event);
void publish_user(const event_t& event);
}} // end namespace

namespace std { // NOLINT
template<>
struct hash<server::registry> {
    auto operator()(const server::registry& key) const {
        return key.hash();
    }
};
} // end namespace

/*!
 * Coventry registry state and access.
 * \file registry.hpp
 */
#endif
