/*
 * Copyright (C) 2022-2023 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "compiler.hpp"
#include "ipc.hpp"

#include <digest.hpp>
#include <strings.hpp>
#include <print.hpp>
#include <keyfile.hpp>
#include <args.hpp>

#include <unistd.h>
#include <termios.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/stat.h>

using namespace crypto;

namespace {
args::option opt_config('c',"--config",     "use config file", "path");
args::option opt_digest('d',"--digest",     "digest(s) to use", "hash");
args::option opt_realm('r', "--realm",      "manually set realm", "domain");
args::option helpflag('h',  "--help",       "display this list");
args::option althelp('?',   nullptr, nullptr);

auto sha256_digest(std::string_view userid, std::string_view realm, std::string_view secret) {
    digest_t sha(EVP_sha256());
    sha.update(userid.data(), userid.size());
    sha.update(":", 1);
    sha.update(realm.data(), realm.size());
    sha.update(":", 1);
    sha.update(secret.data(), secret.size());
    sha.finish();
    return to_hex(sha.data(), sha.size());
}

auto md5_digest(std::string_view userid, std::string_view realm, std::string_view secret) {
    digest_t md5(EVP_md5());
    md5.update(userid.data(), userid.size());
    md5.update(":", 1);
    md5.update(realm.data(), realm.size());
    md5.update(":", 1);
    md5.update(secret.data(), secret.size());
    md5.finish();
    return to_hex(md5.data(), md5.size());
}

void echo(bool enable = true) {
    struct termios tty{};
    tcgetattr(STDIN_FILENO, &tty);
    if(!enable)
        tty.c_lflag &= ~ECHO;
    else
        tty.c_lflag |= ECHO;

    (void) tcsetattr(STDIN_FILENO, TCSANOW, &tty);
}

void digest(std::string_view user, std::string_view realm, std::string_view algo) {
    if(is(opt_digest))
        algo = lower_case(*opt_digest);

    echo(false);
    std::cout << "password: ";
    std::cout.flush();
    std::string password;
    std::cin >> password;
    std::cout << std::endl;

    std::cout << "verify: ";
    std::cout.flush();
    std::string verify;
    std::cin >> verify;
    echo(true);
    std::cout << std::endl;

    if(verify != password)
        die(5, "passwords do not match");

    print("[{}]\n", user);
    if(algo.find("md5") != std::string_view::npos)
        print("md5 = {}\n", md5_digest(user, realm, password));
    if(algo.find("256") != std::string_view::npos)
        print("sha256 = {}\n", sha256_digest(user, realm, password));
}
} // end namespace

auto main(int argc, const char **argv) -> int { // NOLINT
    atexit([]{
        echo(true);
    });

    try {
        auto count = args::parse(argc, argv);
        if(is(helpflag) || is(althelp) || count != 1) {
            args::help({"pbx-passwd [options] user-id"}, {
                "pbx-passwd " PROJECT_VERSION,
                "David Sugar <tychosoft@gmail.com",
                "Create a coventry password digest for a sip user"
            });
            return 0;
        }

        auto user = args::list()[0];
        if(is(opt_realm)) {
            digest(user, *opt_realm, "md5,sha-256");
            return 0;
        }

        if(is(opt_config)) {
            keyfile config;
            config.load(*opt_config);
            auto& hostname = config["server"]["hostname"];
            auto& realm = config["server"]["realm"];
            auto& algo = config["server"]["algorithm"];

            if(algo.empty())
                algo = "MD5,SHA-256";

            algo = lower_case(algo);
            if(hostname.empty()) {
                char buf[256];
                auto result = gethostname(buf, sizeof(buf));
                if(result != 0)
                    throw std::runtime_error("cannot get hostname");
                buf[sizeof(buf) - 1] = 0;
                hostname = buf;
            };
            //net_fqdn(hostname);
            if(realm.empty())
                realm = hostname;
            digest(user, realm, algo);
            return 0;
        }

        auto rshm = shm_open(SERVER_IPCPATH ".registry", O_RDONLY, 0660);
        if(rshm < 0)
            throw std::runtime_error("failed");
        struct stat rino{};
        if(fstat(rshm, &rino) < 0)
            throw std::runtime_error("offline");
        auto rmax = (rino.st_size - sizeof(pbx_sys)) / sizeof(pbx_reg);
        auto rmem = mmap(nullptr, sizeof(pbx_reg) * rmax + sizeof(pbx_sys), PROT_READ, MAP_SHARED, rshm, 0);
        auto reg = static_cast<pbx_reg*>(rmem);
        auto sys = reinterpret_cast<const pbx_sys *>(&reg[rmax]);
        digest(user, sys->realm, sys->digest);
    }
    catch(const bad_arg& e) {
        die(-1, "Usage: pbx-passwd [options] id\nerror: {}\n", e.what());
    }
    catch(const std::exception& e) {
        die(-1, "error: {}\n", e.what());
    }
    catch(...) {
        die(-1, "error: unknown error\n");
    }
    return 0;
}
