/*
 * Copyright (C) 2022-2023 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "compiler.hpp"
#include "ipc.hpp"

#include <filesystem.hpp>
#include <print.hpp>
#include <args.hpp>

#include <fcntl.h>
#include <mqueue.h>

namespace {
args::option helpflag('h',  "--help",       "display this list");
args::option quiet('q',     "--quiet",      "suppress error output");
args::option althelp('?',   nullptr, nullptr);

class ipc_error final : public std::exception {
public:
    explicit ipc_error(const char *text, int code) noexcept : msg(text), err(code) {}

    auto what() const noexcept -> const char * override {
        return msg;
    }

    auto exit() const noexcept {
        return err;
    }

private:
    const char *msg;
    int err;
};
} // end namespace

auto main(int argc, const char **argv) -> int { // NOLINT
    auto exit_code = 0;
    try {
        auto count = args::parse(argc, argv);
        if(is(helpflag) || is(althelp) || count > 1) {
            args::help({"pbx-runstate [new-state]"}, {
                "pbx-runstate " PROJECT_VERSION,
                "David Sugar <tychosoft@gmail.com",
                "Change pbx running mode"
            });
            return 0;
        }

        fsys::current_path(SERVER_PREFIX_PATH);
        if(!count) {
            std::error_code error;
            auto symlink = fsys::read_symlink("state.conf", error);
            if(!symlink.empty() && fsys::is_regular_file(symlink)) {
                const auto filename = symlink.filename().string();
                const auto path = symlink.string();
                if (filename[0] != '.' && path == filename) {
                    print("{}\n", symlink.stem().string());
                    ::exit(0);
                }
            }
            print("default\n");
            ::exit(0);
        }

        auto state = std::string{args::list()[0]};
        if(state == "default") {
            fsys::remove("state.conf");
        }
        else {
            state += ".state";
            if(!fsys::is_regular_file(state))
                throw std::runtime_error(print_format("{}: unknown state", state));
            fsys::remove("state.conf");
            fsys::create_symlink(state, "state.conf");
        }

        pbx_msg msg{.type = PBX_RELOAD, .ver = PBX_VERSION};
        auto mq = mq_open(SERVER_IPCPATH, O_WRONLY, 0660, nullptr);
        if(mq == MQD_INVALID)
            ::exit(0);
        if(mq_send(mq, reinterpret_cast<const char *>(&msg), sizeof(msg), 0) < 0)
            throw ipc_error("send failed", 4);
    }
    catch(const bad_arg& e) {
        die(-1, "Usage: pbx-runstate [new-state]\nerror: {}\n", e.what());
    }
    catch(const ipc_error& e) {
        exit_code = e.exit();
        if(!is(quiet))
            die(exit_code, "error: {}\n", e.what());
    }
    catch(const std::exception& e) {
        if(is(quiet))
            exit_code = -1;
        else
            die(-1, "error: {}\n", e.what());
    }
    catch(...) {
        if(is(quiet))
            exit_code = -1;
        else
            die(-1, "error: unknown error\n");
    }
    return exit_code;
}
